'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var React = require('react');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var React__default = /*#__PURE__*/_interopDefaultLegacy(React);

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __read(o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
}

function __spread() {
    for (var ar = [], i = 0; i < arguments.length; i++)
        ar = ar.concat(__read(arguments[i]));
    return ar;
}

function tween(params, t) {
    if (params.fixed)
        return params.value;
    var _a = __read(params.interval, 2), start = _a[0], end = _a[1];
    var _b = __read(params.extremes, 2), from = _b[0], to = _b[1];
    if (t < start)
        return from;
    if (t > end)
        return to;
    var x = (t - start) / (end - start);
    var ease = params.ease || easing.linear;
    return from + ease(x) * (to - from);
}
function stagger(_a, index, count) {
    var _b = __read(_a, 2), start = _b[0], end = _b[1];
    if (count <= 1)
        return [start, end];
    var totalDuration = end - start;
    var stepDuration = totalDuration / Math.pow(count, 1 / 8);
    var tick = (totalDuration - stepDuration) / (count - 1);
    var stepStart = start + tick * index;
    var stepEnd = stepStart + stepDuration;
    return [stepStart, stepEnd];
}
var easing = {
    linear: function (t) {
        return t;
    },
    easeInQuad: function (t) {
        return t * t;
    },
    easeOutQuad: function (t) {
        return t * (2 - t);
    },
    easeInOutCubic: function (t) {
        return t < 0.5
            ? 4 * t * t * t
            : (t - 1) * (2 * t - 2) * (2 * t - 2) + 1;
    },
};

function useLineTransitions(prevLines, nextLines) {
    return React.useMemo(function () { return getLineTransitions(prevLines, nextLines); }, [prevLines, nextLines]);
}
function getLineTransitions(prevLines, nextLines) {
    var sortedKeys = sortUniqueConcat(prevLines.map(function (l) { return l.key; }), nextLines.map(function (l) { return l.key; }));
    var enterIndex = 0;
    var exitIndex = 0;
    var linesData = sortedKeys.map(function (key) {
        var prevIndex = prevLines.findIndex(function (l) { return l.key === key; });
        var nextIndex = nextLines.findIndex(function (l) { return l.key === key; });
        var prevLine = prevLines[prevIndex];
        var nextLine = nextLines[nextIndex];
        return {
            key: key,
            element: (prevLine === null || prevLine === void 0 ? void 0 : prevLine.element) || nextLine.element,
            prevIndex: !prevLine ? null : prevIndex,
            nextIndex: !nextLine ? null : nextIndex,
            enterIndex: !prevLine ? enterIndex++ : null,
            exitIndex: !nextLine ? exitIndex++ : null,
            elementWithProgress: (prevLine === null || prevLine === void 0 ? void 0 : prevLine.elementWithProgress) || (nextLine === null || nextLine === void 0 ? void 0 : nextLine.elementWithProgress),
        };
    });
    var enterCount = enterIndex;
    var exitCount = exitIndex;
    return {
        lines: linesData.map(function (lineData) {
            return getLineTransition(lineData, enterCount, exitCount);
        }),
        verticalInterval: verticalInterval(enterCount, exitCount),
    };
}
function getLineTransition(_a, enterCount, exitCount) {
    var element = _a.element, key = _a.key, prevIndex = _a.prevIndex, nextIndex = _a.nextIndex, enterIndex = _a.enterIndex, exitIndex = _a.exitIndex, elementWithProgress = _a.elementWithProgress;
    // startY is the progress when we start moving vertically
    // endY is when we stop
    var _b = __read(verticalInterval(enterCount, exitCount), 2), startY = _b[0], endY = _b[1];
    if (prevIndex == null) {
        // entering line
        return {
            element: element,
            elementWithProgress: elementWithProgress,
            key: key,
            tweenY: { fixed: true, value: nextIndex },
            tweenX: {
                fixed: false,
                extremes: [1, 0],
                ease: easing.easeOutQuad,
                interval: stagger([endY, 1], enterIndex, enterCount),
            },
        };
    }
    if (nextIndex == null) {
        // exiting line
        return {
            element: element,
            elementWithProgress: elementWithProgress,
            key: key,
            tweenY: { fixed: true, value: prevIndex },
            tweenX: {
                fixed: false,
                extremes: [0, -1],
                ease: easing.easeInQuad,
                interval: stagger([0, startY], exitIndex, exitCount),
            },
        };
    }
    return {
        element: element,
        elementWithProgress: elementWithProgress,
        key: key,
        tweenY: {
            fixed: false,
            extremes: [prevIndex, nextIndex],
            interval: [startY, endY],
            ease: easing.easeInOutCubic,
        },
        tweenX: { fixed: true, value: 0 },
    };
}
function sortUniqueConcat(a, b) {
    return __spread(new Set(a.concat(b))).sort(function (x, y) { return x - y; });
}
function verticalInterval(enterCount, exitCount) {
    if (enterCount <= 0 && exitCount <= 0)
        return [0, 1];
    if (enterCount <= 0 && exitCount >= 1)
        return [0.33, 1];
    if (enterCount >= 1 && exitCount <= 0)
        return [0, 0.67];
    return [0.25, 0.75];
}

function Lines(_a) {
    var lines = _a.lines, prevFocusKeys = _a.prevFocusKeys, nextFocusKeys = _a.nextFocusKeys, focusWidth = _a.focusWidth, lineHeight = _a.lineHeight, progress = _a.progress;
    return (React__default['default'].createElement(React__default['default'].Fragment, null, lines.map(function (_a) {
        var element = _a.element, key = _a.key, tweenX = _a.tweenX, tweenY = _a.tweenY, elementWithProgress = _a.elementWithProgress;
        var dx = tween(tweenX, progress);
        var dy = tween(tweenY, progress);
        var opacity = getOpacity(prevFocusKeys, key, nextFocusKeys, progress, dx);
        return (React__default['default'].createElement(LineContainer, { key: key, dx: dx * focusWidth, dy: dy * lineHeight, opacity: opacity, width: focusWidth }, elementWithProgress
            ? elementWithProgress(progress)
            : element));
    })));
}
function LineContainer(_a) {
    var children = _a.children, dx = _a.dx, dy = _a.dy, opacity = _a.opacity, width = _a.width;
    return (React__default['default'].createElement("div", { style: {
            position: "absolute",
            top: 0,
            left: 0,
            transform: "translate(" + dx + "px, " + dy + "px)",
            opacity: opacity,
            width: width,
            display: opacity <= 0 ? "none" : undefined,
        } }, children));
}
var OFF_OPACITY = 0.33;
function getOpacity(prevFocusKeys, key, nextFocusKeys, progress, dx) {
    return (tween({
        fixed: false,
        extremes: [
            prevFocusKeys.includes(key) ? 0.99 : OFF_OPACITY,
            nextFocusKeys.includes(key) ? 0.99 : OFF_OPACITY,
        ],
        interval: [0, 1],
    }, progress) -
        Math.abs(dx) * 1);
}

function SmoothLines(_a) {
    var progress = _a.progress, containerHeight = _a.containerHeight, containerWidth = _a.containerWidth, lineHeight = _a.lineHeight, lineWidth = _a.lineWidth, _b = _a.prevLines, prevLines = _b === void 0 ? [] : _b, _c = _a.nextLines, nextLines = _c === void 0 ? [] : _c, prevFocus = _a.prevFocus, nextFocus = _a.nextFocus, center = _a.center, _d = _a.minZoom, minZoom = _d === void 0 ? 0 : _d, _e = _a.maxZoom, maxZoom = _e === void 0 ? 1.2 : _e;
    var _f = useLineTransitions(prevLines, nextLines), lines = _f.lines, verticalInterval = _f.verticalInterval;
    var prevFocusKeys = prevFocus.map(function (index) { var _a; return (_a = prevLines[index]) === null || _a === void 0 ? void 0 : _a.key; });
    var nextFocusKeys = nextFocus.map(function (index) { var _a; return (_a = nextLines[index]) === null || _a === void 0 ? void 0 : _a.key; });
    var _g = __read(getContentProps({
        containerWidth: containerWidth,
        containerHeight: containerHeight,
        lineWidth: Array.isArray(lineWidth)
            ? lineWidth[0]
            : lineWidth,
        lineHeight: lineHeight,
        minZoom: minZoom,
        maxZoom: maxZoom,
        horizontalCenter: !!center,
        focusLineIndexList: prevFocus,
        originalContentHeight: prevLines.length * lineHeight,
    }), 5), prevZoom = _g[0], prevDX = _g[1], prevDY = _g[2], prevContentHeight = _g[3], prevContentWidth = _g[4];
    var _h = __read(getContentProps({
        containerWidth: containerWidth,
        containerHeight: containerHeight,
        lineWidth: Array.isArray(lineWidth)
            ? lineWidth[1]
            : lineWidth,
        lineHeight: lineHeight,
        minZoom: minZoom,
        maxZoom: maxZoom,
        horizontalCenter: !!center,
        focusLineIndexList: nextFocus,
        originalContentHeight: nextLines.length * lineHeight,
    }), 5), nextZoom = _h[0], nextDX = _h[1], nextDY = _h[2], nextContentHeight = _h[3], nextContentWidth = _h[4];
    var zoom = tweenProp(prevZoom, nextZoom, progress);
    var dx = tweenProp(prevDX, nextDX, progress);
    var dy = tweenProp(prevDY, nextDY, progress, verticalInterval);
    var focusHeight = tweenProp(prevContentHeight, nextContentHeight, progress);
    var focusWidth = tweenProp(prevContentWidth, nextContentWidth, progress);
    return (React__default['default'].createElement(Container, { width: containerWidth, height: containerHeight },
        React__default['default'].createElement(Content, { dx: dx, dy: dy, scale: zoom, height: Math.max(focusHeight, containerHeight), width: Math.max(focusWidth, containerWidth) },
            React__default['default'].createElement(Lines, { lines: lines, prevFocusKeys: prevFocusKeys, nextFocusKeys: nextFocusKeys, focusWidth: focusWidth, lineHeight: lineHeight, progress: progress }))));
}
function getContentProps(_a) {
    var containerWidth = _a.containerWidth, containerHeight = _a.containerHeight, lineWidth = _a.lineWidth, lineHeight = _a.lineHeight, minZoom = _a.minZoom, maxZoom = _a.maxZoom, focusLineIndexList = _a.focusLineIndexList, originalContentHeight = _a.originalContentHeight, horizontalCenter = _a.horizontalCenter;
    var extremes = [
        Math.min.apply(Math, __spread(focusLineIndexList)),
        Math.max.apply(Math, __spread(focusLineIndexList)),
    ];
    var originalFocusHeight = (extremes[1] - extremes[0] + 3) * lineHeight;
    var zoom = Math.max(Math.min(containerWidth / lineWidth, containerHeight / originalFocusHeight, maxZoom), minZoom);
    var contentHeight = originalContentHeight * zoom;
    var focusStart = (extremes[0] - 1) * lineHeight * zoom;
    var focusEnd = (extremes[1] + 2) * lineHeight * zoom;
    var focusCenter = (focusEnd + focusStart) / 2;
    var focusHeight = focusEnd - focusStart;
    var dy = containerHeight > contentHeight
        ? (containerHeight - contentHeight) / 2
        : clamp(containerHeight / 2 - focusCenter, Math.max(containerHeight - contentHeight, -focusStart // to ensure first focus line is shown when focus is bigger than container
        ), 0);
    var dx = horizontalCenter
        ? Math.max(containerWidth / 2 - (lineWidth * zoom) / 2, 0)
        : 0;
    return [
        zoom,
        dx,
        dy,
        focusHeight,
        lineWidth * zoom,
    ];
}
function Container(_a) {
    var width = _a.width, height = _a.height, children = _a.children;
    return (React__default['default'].createElement("div", { style: {
            width: width,
            height: height,
            position: "relative",
        } }, children));
}
function Content(_a) {
    var dx = _a.dx, dy = _a.dy, scale = _a.scale, height = _a.height, width = _a.width, children = _a.children;
    return (React__default['default'].createElement("div", { style: {
            position: "absolute",
            top: 0,
            left: 0,
            transformOrigin: "top left",
            width: width,
            height: height,
            overflow: "hidden",
        } },
        React__default['default'].createElement("div", { style: {
                position: "absolute",
                top: 0,
                left: 0,
                transform: "translateX(" + dx + "px) translateY(" + dy + "px) scale(" + scale + ")",
                transformOrigin: "top left",
            } }, children)));
}
function tweenProp(start, end, progress, interval) {
    if (interval === void 0) { interval = [0, 1]; }
    return tween({
        fixed: false,
        interval: interval,
        extremes: [start, end],
        ease: easing.easeInOutCubic,
    }, progress);
}
function clamp(num, min, max) {
    return num <= min ? min : num >= max ? max : num;
}

exports.SmoothLines = SmoothLines;
